#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "ATmega88PA.h"
#include "device.h"
#include "onewire.h"
#include "stub.h"
#include "ds18B20.h"

#define SKIP_ROM 0xCC
#define CONVERT_TEMPERATURE 0x44
#define WRITE_SCRATCHPAD 0x4E
#define READ_SCRATCHPAD 0xBE
#define COPY_SCRATCHPAD 0x48
#define TEMP_LSB 0xFF
#define TEMP_MSB 0xFF
#define TH_USER_1 0xFF
#define TL_USER_2 0xFF
#define CONFIG 0xFF
#define CONFIG_9BIT 0x1F
#define CONFIG_10BIT 0x3F
#define CONFIG_11BIT 0x5F
#define CONFIG_12BIT 0x7F
#define RESERVED 0xFF
#define CRC 0xFF
#define CONVERSION_COMPLETE 0xFF

#define REPEATER_ON TIMER0_COMPARE_MATCH_A_INTERRUPT_ENABLE;
#define REPEATER_OFF TIMER0_COMPARE_MATCH_A_INTERRUPT_DISABLE;

typedef enum {
    CONFIG_SEND,
    CONFIG_SAVE,
    CONVERT_RESET,
    CONVERT,
    CONVERT_CHECK,
    READ
} DS18B20_STATE;

static DS18B20_STATE state;
static uint8_t buffer[11];
static uint8_t error_counter = 0;

static const uint8_t crc_table[256] PROGMEM = {
    0, 94, 188, 226, 97, 63, 221, 131, 194, 156, 126, 32, 163, 253, 31, 65,
	157, 195, 33, 127, 252, 162, 64, 30, 95, 1, 227, 189, 62, 96, 130, 220,
	35, 125, 159, 193, 66, 28, 254, 160, 225, 191, 93, 3, 128, 222, 60, 98,
	190, 224, 2, 92, 223, 129, 99, 61, 124, 34, 192, 158, 29, 67, 161, 255,
	70, 24, 250, 164, 39, 121, 155, 197, 132, 218, 56, 102, 229, 187, 89, 7,
	219, 133, 103, 57, 186, 228, 6, 88, 25, 71, 165, 251, 120, 38, 196, 154,
	101, 59, 217, 135, 4, 90, 184, 230, 167, 249, 27, 69, 198, 152, 122, 36,
	248, 166, 68, 26, 153, 199, 37, 123, 58, 100, 134, 216, 91, 5, 231, 185,
	140, 210, 48, 110, 237, 179, 81, 15, 78, 16, 242, 172, 47, 113, 147, 205,
	17, 79, 173, 243, 112, 46, 204, 146, 211, 141, 111, 49, 178, 236, 14, 80,
	175, 241, 19, 77, 206, 144, 114, 44, 109, 51, 209, 143, 12, 82, 176, 238,
	50, 108, 142, 208, 83, 13, 239, 177, 240, 174, 76, 18, 145, 207, 45, 115,
	202, 148, 118, 40, 171, 245, 23, 73, 8, 86, 180, 234, 105, 55, 213, 139,
	87, 9, 235, 181, 54, 104, 138, 212, 149, 203, 41, 119, 244, 170, 72, 22,
	233, 183, 85, 11, 136, 214, 52, 106, 43, 117, 151, 201, 74, 20, 246, 168,
	116, 42, 200, 150, 21, 75, 169, 247, 182, 232, 10, 84, 215, 137, 107, 53
};

static uint8_t DS18B20_CheckCRC() {
    uint8_t crc = 0;
    for(uint8_t i = 2; i < 11; i++) {
        crc ^= buffer[i];
        crc = pgm_read_byte(&(crc_table[crc]));
    }
    if(crc) { return 0; }
    return 1;
}

int16_t DS18B20_GetTemperature(void) {
    int16_t temp = (((buffer[3]) << 8) | buffer[2]) << 4;
    return temp;
}

static uint8_t DS18B20_ConvertTemperature(void) {
    buffer[0] = SKIP_ROM;
    buffer[1] = CONVERT_TEMPERATURE;
    return ONEWIRE_SendData(buffer, 2);
}

static void DS18B20_CheckConversion(void) {
    buffer[0] = 0xFF;
    ONEWIRE_SendData(buffer, 1);
}

static uint8_t DS18B20_WriteScratchpad(void) {
    buffer[0] = SKIP_ROM;
    buffer[1] = WRITE_SCRATCHPAD;
    buffer[2] = TH_USER_1;
    buffer[3] = TL_USER_2;
    buffer[4] = CONFIG_9BIT;
    return ONEWIRE_SendData(buffer, 5);
}

static uint8_t DS18B20_ReadScratchpad(void) {
    buffer[0] = SKIP_ROM;
    buffer[1] = READ_SCRATCHPAD;
    buffer[2] = TEMP_LSB;
    buffer[3] = TEMP_MSB;
    buffer[4] = TH_USER_1;
    buffer[5] = TL_USER_2;
    buffer[6] = CONFIG;
    buffer[7] = RESERVED;
    buffer[8] = RESERVED;
    buffer[9] = RESERVED;
    buffer[10] = CRC;
    return ONEWIRE_SendData(buffer, 11);
}

static uint8_t DS18B20_CopyScratchpad(void) {
    buffer[0] = SKIP_ROM;
    buffer[1] = COPY_SCRATCHPAD;
    return ONEWIRE_SendData(buffer, 2);
}

static void DS18B20_Config(void) {
    ONEWIRE_SendReset();
    state = CONFIG_SEND;
}

void DS18B20_ReadTemperature(void) {
    ONEWIRE_SendReset();
    state = CONVERT_RESET;
    error_counter = 0;
    REPEATER_ON;
}

static int8_t counter = 0;

static void DS18B20_DevicePresent(void) {
    counter = 0;
    if(state == CONVERT_RESET) {
        state = CONVERT;
    }
    REPEATER_ON;
}

static void DS18B20_ReceiveComplete(void) {
    switch(state) {
        case CONFIG_SEND:
            REPEATER_OFF;
            ONEWIRE_SendReset();
            state = CONFIG_SAVE;
            break;
        case CONFIG_SAVE:
            REPEATER_OFF;
            break;
        case CONVERT_CHECK:
            if(buffer[0] == CONVERSION_COMPLETE) {
                ONEWIRE_SendReset();
                state = READ;
                REPEATER_OFF;
            }
            break;
        case READ:
            if(DS18B20_CheckCRC()) {
                REPEATER_OFF;
                if(buffer[6] & 0x60) {
                    DS18B20_Config();
                }
                DS18B20_ReadComplete();
            } else {
                //REPEATER_ON;
            }
            break;
        default: break;
    }
}

void DS18B20_Activate(void) {
    ONEWIRE_Init();
    ONEWIRE_DevicePresent = DS18B20_DevicePresent;
    ONEWIRE_ReceiveComplete = DS18B20_ReceiveComplete;
    ONEWIRE_Activate();
}

static void DS18B20_Deactivate(void) {
    REPEATER_OFF;
}

void DS18B20_Init(void) {
    DS18B20_ReadComplete = STUB_void;
    DS18B20_ReadError = STUB_void;
    TIMER0_PWMA_NORMAL_PORT_OPERATION;
    TIMER0_COMPARE_MATCH_A_VALUE(0x7F);
    TIMER0_CLEAR_COMPARE_MATCH_A_INTERRUPT_FLAG;
    DS18B20_Deactivate();
}

//Repeater
ISR(TIMER0_COMPA_vect) {
    if(--counter > 0) { return; }
    counter = 80;
    switch(state) {
        case CONFIG_SEND:
            if(DS18B20_WriteScratchpad()) {
                REPEATER_OFF;
            }
            break;
        case CONFIG_SAVE:
            if(DS18B20_CopyScratchpad()) {
                REPEATER_OFF;
            }
            break;
        case CONVERT_RESET:
            if(++error_counter > 5) {
                REPEATER_OFF;
                DS18B20_ReadError();
            }
            break;
        case CONVERT:
            if(DS18B20_ConvertTemperature()) {
                state = CONVERT_CHECK;
            }
            break;
        case CONVERT_CHECK: DS18B20_CheckConversion(); break;
        case READ:
            if(DS18B20_ReadScratchpad()) {
                REPEATER_OFF;
            }
            break;
    }
}
